@@ -1,75 +1,15 @@
 from __future__ import annotations
 
-from functools import total_ordering
-from itertools import product
-from typing import Hashable, FrozenSet, Callable, Union
+from typing import Callable, Hashable
 
 import attr
 import funcy as fn
 
-
-def _freeze(alphabet):
-    if alphabet is None or isinstance(alphabet, SupAlphabet):
-        return SupAlphabet()
-    elif isinstance(alphabet, ProductAlphabet):
-        return alphabet
-    return frozenset(alphabet)
-
-
-@attr.s(frozen=True, auto_attribs=True, eq=False)
-@total_ordering
-class SupAlphabet:
-    """Alphabet containing all other alphabets."""
-    def __hash__(self):
-        return hash(id(self))
-
-    def __contains__(self, _):
-        return True
-
-    def __lt__(self, other):
-        return False
-
-    def __eq__(self, other):
-        return isinstance(other, SupAlphabet)
-
-
-@attr.s(frozen=True, auto_attribs=True, repr=False, eq=False)
-@total_ordering
-class ProductAlphabet:
-    """Implicit encoding of product alphabet."""
-    left: Alphabet = attr.ib(converter=_freeze)
-    right: Alphabet = attr.ib(converter=_freeze)
-
-    def __hash__(self):
-        return hash((self.left, self.right))
-
-    def __eq__(self, other):
-        return (self.left == other.left) and (self.right == other.right)
-
-    def __lt__(self, other):
-        return (self.left < other.left) and (self.right < other.right)
-
-    def __contains__(self, elem):
-        assert len(elem) == 2
-        return (elem[0] in self.left) and (elem[1] in self.right)
-
-    def __iter__(self):
-        return product(self.left, self.right)
-
-    def __repr__(self):
-        left, right = map(
-            lambda x: x if isinstance(x, SupAlphabet) else set(x),
-            [self.left, self.right]
-        )
-        return f"{left} x {right}"
-
-    def __len__(self):
-        return len(self.left) * len(self.right)
+from dfa.alphabet import Letter, Alphabet, ProductAlphabet, SupAlphabet
+from dfa.alphabet import _freeze
 
 
 State = Hashable
-Letter = Hashable
-Alphabet = Union[FrozenSet[Letter], ProductAlphabet, SupAlphabet]
 
 
 @attr.s(frozen=True, auto_attribs=True)
@@ -86,27 +26,27 @@ class DFA:
     inputs: Alphabet = attr.ib(converter=_freeze, default=SupAlphabet())
     outputs: Alphabet = attr.ib(converter=_freeze, default={True, False})
 
-    def run(self, *, start=None):
+    def run(self, *, start=None, label=False):
         """Creates a co-routine that simulates the automata:
             - Takes in characters from the input alphabet.
             - Yields states.
         """
         state = self.start if start is None else start
+        labeler = self._label if label else lambda x: x
 
         while True:
-            char = yield state
+            char = yield labeler(state)
             assert (self.inputs is None) or (char in self.inputs)
             state = self._transition(state, char)
 
-    def trace(self, word, *, start=None):
+    def trace(self, word, *, start=None, label=False):
         """Creates a generator that yields the sequence of states that
         results from reading the word.
         """
-        machine = self.run(start=start)
+        machine = self.run(start=start, label=label)
 
         for char in fn.concat([None], word):
-            state = machine.send(char)
-            yield state
+            yield machine.send(char)
 
     def transition(self, word, *, start=None):
         """Returns the state the input word accesses."""
@@ -114,13 +54,13 @@ class DFA:
 
     def label(self, word, *, start=None):
         """Returns the label of the input word."""
-        output = self._label(self.transition(word, start=start))
+        output = fn.last(self.trace(word, start=start, label=True))
         assert (self.outputs is None) or (output in self.outputs)
         return output
 
     def transduce(self, word, *, start=None):
         """Returns the labels of the states traced by the input word."""
-        return tuple(map(self._label, self.trace(word, start=start)))[:-1]
+        return tuple(self.trace(word, start=start, label=True))[:-1]
 
     @fn.memoize()
     def states(self):
